// Fichier main.c
/////////////////////////////////////////////////////////

/****************************************************************
 *                                                    			*
 * .Date de cration : 11/01/08                         		*
 * .Auteur : Guibert David                              		*       
 *           Cadic Matthieu										*
 *           Luzeau Yohan                                		*
 * .Descriptions :                                      		*
 *      - Fichier principal pour la gestion du rouleau droit   	*
 *                                                      		*
 *                                                      		*
 *                                                      		*
 * .Remarques :                                         		*
 *      - Gestion des ordre reus par interruption plutt		*
 *		  que par polling								   		*
 *      - Adresse de la carte droite 0x4E          				*
 * 		- Transmission IC sur interruption (priorit haute)	*
 * 		- Timer sur interruption (priorit basse)               *
 *		- La led rouge marque l'interruption timer				*
 *		- La led ambre marque une transmission IC				*
 *																*
 ****************************************************************/




// ------------ Directives de compilation
#pragma config WDT = OFF	// Pas de  watchdog
#pragma config OSC = HS		// Quartz HS pour un quartz 20MHz
#pragma config LVP = OFF	// Pas de Low Voltage Programming


// --------------------------------------------------------------------------
// Enttes ncessaires
#include <p18f252.h>		// Pour les registres du PIC18F252
#include <delays.h>    		// Pour les temporisations
#include <timers.h>			// Pour la base de temps de 10ms
#include <adc.h>			// Bibliothque des convertisseurs Analogiques Numriques
#include <usart.h>			// Pour la liaison srie
#include <stdio.h>


#include "Moteur.h"			// Librairie de gestion du moteur
#include "I2C_Slave.h"		// Librairie pour l'I2C
#include "led.h"			// Librairie pour les leds
#include "codeur.h"			// Librairie pour le codeur


// --------------------------------------------------------------------------
// Constantes
#define 	T_TIMER3 		40535		   	// 5 ms avant dbordement du Timer
											// Priode du quartz : 1/20e6 = 50ns
											// Priode  l'entre du prescaler : 50ns.4 = 200ns
											// Priode  la sortie du prescaler : 200ns.1 = 200ns
											// Nombre de priodes avant dbordement du timer : 10000 incrments
											// Priode entre deux interruptions : 5000*200ns = 1ms
											// Priode de 1ms -> 1KHz
											// 60536 pour 1ms
											// 55536 pour 2ms
											// 40535 pour 5ms
											// 15535 pour 10ms

#define		COMMANDE_MAX	1023			// Commande max autorise sur les moteurs
//#define 	Ki				0.0231902375	// Coefficient intgral			
// XM, 08/05
#define 	Ki				0.04	// Coefficient intgral			
#define 	Kp				2.16349043		// Coefficient proportionnel

// --------------------------------------------------------------------------
// Variables partage
volatile int 			Speed_Shared=0;					// Vitesse mesure au codeurs
volatile int 			Consigne_Shared=0;				// Consigne dsire
volatile unsigned char 	CompteurLedAmbre_Shared=0;		// Compteur pour allumer la led ambre sur un dialogue IC

#define SPEEDS_COUNT	100
volatile int			speeds_avg=0;

void Init_Speeds(void){
	int i;
}

int abs( int val ){
	return val>0 ? val : -val;
}

void Update_Speeds(int new_speed){
	static int		speeds_init=1;			// permet d'indiquer que c'est la premire fois qu'on passe dans la fonction
	static int		speeds[SPEEDS_COUNT];	// l'historique des vitesses
	static int		speeds_index=0;			// l'index de la vitesse  ajouter
	static long int	speeds_sum=0;			// la somme de toutes les vitesses du tableau
	
	int i, diff;

	
	if( speeds_init ){						// la premire fois, initialiser le tableau  0
		for( i=0; i<SPEEDS_COUNT; i++ )
			speeds[i] = 0;
		speeds_init = 0;
	}
	
	
	diff = new_speed	- speeds[speeds_index];
	speeds_sum += diff;
	speeds[speeds_index] = new_speed;
	speeds_avg = speeds_sum/SPEEDS_COUNT;
	
	if( speeds_index<SPEEDS_COUNT-1 )
		speeds_index++;
	else
		speeds_index = 0;
}
/*
void Update_Speeds(int new_speed){
	static int		index = 0;			// l'index de la vitesse  ajouter

	// si la mesure ne correspond pas  la consigne, annuler la srie ventuellement commence
	if( abs(new_speed-2*Consigne_Shared)>1 )
		index = 0;
	else{
		// incrmenter le nombre de bonnes mesures
		index++;
	}
	
	// 100 bonne mesures atteintes
	if( index>=100 ){
		index = 100;
		speeds_avg = 2*Consigne_Shared;
	}else{
		speeds_avg = 0;
	}
			
}
*/
// --------------------------------------------------------------------------

// Prototype de la routine de service d'interuptions de priorit haute
void Interrupt_High_Priority(void);					// Utilise pour l'IC


// Prototype de la routine de service d'interuptions de priorit basse
void Interrupt_Low_Priority(void);					// Utilise pour le timer




/////////////////////////////////////////////////////////////////////////////
//                                                                         //
//                            Fonction Principale                          //
//                                                                         //
/////////////////////////////////////////////////////////////////////////////

void main (void)
{
	// Configure le micro	
	
	OpenADC (ADC_0ANA_0REF,ADC_INT_OFF);	// Desactive les convertisseurs Analogiques/Numriques
	Moteur_Init();							// Initialise la commande du moteur
	Delay10KTCYx(250);						// Tempo pour permettre au moteur de s'arrter s'il tournait
 	I2C_Init ();							// Initialise l'I2C
 	Led_Init();								// Configure les ports pour les leds
 	Codeur_Init();							// Configure les codeurs
  	

	// Configure la liaison srie  115200 Bauds:
	OpenUSART	(	USART_TX_INT_OFF	&	// Pas d'interruption sur l'mission
					USART_RX_INT_OFF	&	// Interruption sur la rception
					USART_ASYNCH_MODE	&	// Mode asynchrone
					USART_EIGHT_BIT		& 	// Transmission sur 8 bits (1 octet)
					USART_CONT_RX		&	// Reception continue
					USART_BRGH_HIGH		,	// Haute vitesse
					10);					// 10 = (20e6 / (16 * 115200))-1 
	
	// Configuration du timer 3 pour la base de temps
	
	OpenTimer3(	TIMER_INT_ON &				// Autorise les interruption sur le timer 3
				T3_16BIT_RW &				// Compteur 16 bits
				T3_SOURCE_INT &				// Source interne (Quartz Fosc/4)
				T3_PS_1_1 &					// Prescaler 1/1
				T3_SYNC_EXT_OFF);			// Pas de synchronisation externe

	WriteTimer3(T_TIMER3);					// Charge le compteur du timer 3 avec la valeur de T_Timer3
	PIR2bits.TMR3IF=0;						// Timer3 Interrupt Flag = 0 Baisse le flag du timer 3
	IPR2bits.TMR3IP=0;						// Timer 3 dclench sur priorit basse
	RCONbits.IPEN=1;						// Autorise les niveau de priorit sur les interuptions
	
	
	// Autorise les interruptions	
	INTCONbits.PEIE=1;						// Autorise les interruptions de basse priorit
	INTCONbits.GIE=1;						// Autorise les interruptions de haute priorit

	stdout = _H_USART;
	while(1);								// Rien a faire, tout est gr dans la routine d'interruption
}






/////////////////////////////////////////////////////////////////////////////
//                                                                         //
//                            Interruptions                                // 
//                                                                         // 
/////////////////////////////////////////////////////////////////////////////



//----------------------------------------------------------------------------
// Vecteur d'interruptions de priorit haute (utilis pour l'IC)
#pragma code highVector=0x08						// Vecteur d'interruption haut
void atInterruptHigh(void)
{
	_asm goto Interrupt_High_Priority _endasm		// On saute  la fonction MyHighInterrupt
}
#pragma code 										// Retour  la zone de code




//----------------------------------------------------------------------------
// Vecteur d'interruptions de priorit basse (utilis pour le Timer)
#pragma code lowVector=0x18							// Vecteur d'interruption de priorit basse
void atInterruptLow(void)
{
	_asm goto Interrupt_Low_Priority _endasm		// On saute  la fonction MyHighInterrupt
}
#pragma code 										// Retour  la zone de code




//----------------------------------------------------------------------------
// Routine de service de l'interruption haute priorite
#pragma interrupt Interrupt_High_Priority 
void Interrupt_High_Priority(void)
{
	
	if (PIR1bits.SSPIF) 										// C'est le MSSP qui a lev l'interruption, c'est pour l'IC
	{
		CompteurLedAmbre_Shared=50;								// Compteur (qui allume la led pendant 50ms)		
		PIR1bits.SSPIF = 0; 									// On rautorise l'interruption		
		if (speeds_avg> 1023)	speeds_avg = 1023;				// Borne la vitesse entre [0;+1023]
		if (speeds_avg< 0) 		speeds_avg= 0;					// Pour l'envoie par IC
		Consigne_Shared=InterruptI2C((unsigned char)(speeds_avg>>1));			// On appelle la routine I2C pour traiter la requete				
	}	
}




//----------------------------------------------------------------------------
// Routine de service de l'interruption haute priorite
#pragma interruptlow Interrupt_Low_Priority 
void Interrupt_Low_Priority(void)
{
	
	int 			Speed;								// Vitesse mesure au codeur
	float 			Commande=0;							// Commande applique au moteur
	static float 	CommandeOld=0;						// Commande applique  l'chantillon prcdent
	float 			Error=0;							// Erreur courante en vitesse
	static float 	ErrorOld=0;							// Erreur  l'chantillon prcdent
	int 			Consigne=0;							// Consigne  appliquer sur le moteur
	
	
	
		
	
	
		
	if(PIR2bits.TMR3IF)											// Vrifie qu'il s'agit bien de l'interruption du Timer 3
	{									
		Led_Rouge(1);											// Pour marquer le dbut de l'interruption
		
		
		WriteTimer3(T_TIMER3);									// Recharge le compteur du Timer 3		
		PIR2bits.TMR3IF=0;										// Rautotise l'interruption


		Speed=Codeur_Vitesse();									// Lecture de la vitesse du moteur
		Update_Speeds( Speed );									// met  jour l'historique des vitesses et calcule la vitesse moyenne		
		
		PIE1bits.SSPIE=0;										// Interdit l'interruption IC pour permettre le partage des variables
		Speed_Shared=Speed;										// Tranfert la vitesse mesure dans la variable partage
		Consigne=Consigne_Shared;								// Rcupre la consigned
		Consigne*=2;											// Multiplie la ocnsigne par 2 pour la mettre  l'chelle
		
		PIE1bits.SSPIE=1;										// Autorise de nouveau les interruptions
		
		Error=(float)(Consigne-Speed);								// Calcul l'erreur en vitesse
		Commande=CommandeOld+(Ki-Kp)*ErrorOld+Kp*Error;			// Calcule la commande  appliquer au moteur
		
		ErrorOld=Error;											// Mmorise l'erreur pour le coup suivant

		if (Commande> COMMANDE_MAX) Commande= COMMANDE_MAX;		// Bride la commande du moteur
		if (Commande<-COMMANDE_MAX) Commande=-COMMANDE_MAX;
		CommandeOld=Commande;									// Mmorise la commande pour l'chantillon suivant
		CommandeMoteur ((int)Commande);   						// Envoie la commande sur les moteurs
		
		printf( "%d\t%d\t%d\n", Speed, speeds_avg, 2*Consigne_Shared );
			
		
		if (CompteurLedAmbre_Shared)							// Si le compteur pour la led n'est pas  0
		{
			Led_Ambre(1);										// Allume la led ambre pour signifier qu'il y a une transmission I2C
			CompteurLedAmbre_Shared--;							// Dcrmente le compteur
		}
		else
		{
			Led_Ambre(0);										// Eteind la led
		}
		
		
		Led_Rouge(0);											// Pour marquer la fin de l'interruption
	}
}
	